
#see in glove-win_devc_x64, demo.sh or demo.bat
'''
MacOS and Linux
Go to: https://github.com/stanfordnlp/GloVe

$ git clone http://github.com/stanfordnlp/glove
$ cd glove && make
$ ./demo.sh


Windows 10
https://github.com/anoidgit/GloVe-win

'''
